/*
 * com/ngon/mpeg/MPEGStream.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * This code is based on GPL'd code in MPEG_Play by Joerg Anders
 * which may be found at:
 *
 *     http://rnvs.informatik.tu-chemnitz.de/~jan/MPEG/MPEG_Play.html
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.mpeg;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.InputStream;
import java.io.IOException;

public class MPEGStream {

    private DataInputStream data = null;
    private int bitPos = 0;
    private int bitPosSave = 0;
    private long bitShiftReg = 0L;
    private long bitShiftRegSave = 0L;
    private boolean isEOF = false;

    private long nextBitsValue;
    private int nextBitsAvailable = 0;
    private int remainingBytes = -1;    // # of MPEG layer II bytes 
                                        // remaining. A negative value 
                                        // indicates MPEG layer I.
    private Semaphore videoSemaphore;

    public MPEGStream(InputStream in, Semaphore videoSemaphore) {
        this.data = new DataInputStream(new BufferedInputStream(in));
        this.videoSemaphore = videoSemaphore;
    }

    public void close() throws IOException {
        if (data != null) {
            data.close();
            data = null;
        }
    }

    public boolean isEndOfFile() { 
        return isEOF; 
    }

    // The method skip(long n) skips over n bytes but takes into account that
    // some of them are already in shift register.
    public void skip(long n) {
        if (data == null) {
            return;
        }
        int k = bitPos / 8;
        n -= k;
        bitPos -= k * 8;
        nextBitsAvailable = 0;
        try {
            data.skip(n);
        } catch (IOException e) {
            System.err.println(e.toString());
        }
    }

    // The method getNumReadBytes() tells how many bytes are in 
    // shift register.
    public int getNumReadBytes() {
        if ((bitPos & 0x7) != 0) { // bitPos sould be byte aligned
            System.err.println("getNumReadBytes(): bitPos = " + bitPos);
        }
        return (bitPos / 8);
    }

    public void setNumRemainingBytes(int bytes) {
        remainingBytes = bytes;
    }

    private void fillBitShiftRegister() throws IOException {
        try {
            // If less than 4 bytes remain in layer II data,
            // check whether it's possible to return 32 bits.
            if ((remainingBytes & 0xfffffffc) == 0 ) {
                // remainingBytes > 0 && remainingBytes < 4
                if (remainingBytes * 8 >= 32 - bitPos) { // 32 bits available?
                    while (data != null && 
                           bitPos < 32) { // Yes => read more bytes
                        bitShiftReg = (bitShiftReg << 8) | 
                                      (0xff & data.readByte());
                        bitPos += 8;
                        remainingBytes--;
                    }
                    return;
                }

                // There aren't 32 bits available in layer II data,
                // so save the bits in the bit shift register.
                bitShiftRegSave = bitShiftReg & ((1L << bitPos) - 1L);
                bitPosSave = bitPos;

                // Read the remaining layer II bytes and 
                // put them in the saved shift register.
                for( ; data != null && remainingBytes > 0; remainingBytes--) {
                    bitShiftRegSave <<= 8;
                    bitShiftRegSave |= (0xff & data.readByte());
                    bitPosSave += 8;
                }

                // Prepare for switch to layer I.
                remainingBytes = -1;
                bitPos = 0;
                nextBitsAvailable = 0;
                
                // Suspend layer II thread; wake up layer I thread.
                videoSemaphore.toggle0();
                if (bitPosSave + bitPos > 63) {
                    System.out.println(
                        "Recovery not possible = " + bitPosSave +
                        "; bitPos = " + bitPos);
                }
                // Cut leading bits to "0".
                bitShiftReg &= ((1L << bitPos) - 1L);
                // Recover saved bits.
                bitShiftReg |= (bitShiftRegSave << bitPos);
                nextBitsAvailable = 0;
                bitPos += bitPosSave;
                while (data != null && bitPos < 32) {
                    // Make 32 bits available.
                    bitShiftReg = (bitShiftReg << 8) | 
                                  (0xff & data.readByte());
                    bitPos += 8; 
                    remainingBytes--;
                }
                return;
            }
            else { 
                // Normal processing.
                bitShiftReg = (bitShiftReg << 32) | 
                              (0xffffffffL & data.readInt());
                bitPos += 32;
                remainingBytes -= 4;
            }
        } catch (EOFException e) {
            isEOF = true;
            bitPos += 32;
            bitShiftReg <<= 32;
            System.out.flush();
        }
    }

    // The method getBits() gets the next n bits from shift register 
    // interprets them as integer and returns this integer value.
    public int getBits(int n) throws IOException {
        long val;
        int avail;

        if (nextBitsAvailable >= n) { 
            // If last access was via compareNextBits(), use this result.
            bitPos -= n;
            avail = nextBitsAvailable;
            nextBitsAvailable = 0;
            return (int) nextBitsValue >>> (avail - n);
        }
        nextBitsAvailable = 0;
        if (data != null && bitPos < n) {
            fillBitShiftRegister();
        }
        val = bitShiftReg >>> (bitPos - n);
        val &= (1L << n) - 1L; // Mask leading bits.
        bitPos -= n;
        return (int) val;
    }

    // The method compareNextBits() checks whether the next n bits match the 
    // pattern. If so it returns true; otherwise false. 
    // Note: This method changes the bit pointer physically but not
    // logically.
    public boolean compareNextBits(int n, int pattern) throws IOException {
        nextBitsAvailable = n;
        if (data != null && bitPos < n) {
            fillBitShiftRegister();
        }
        nextBitsValue = bitShiftReg >>> (bitPos - n);
        nextBitsValue &= (1L << n) - 1L; // Mask leading bits.
        return (nextBitsValue == pattern);
    }

    // The method ungetBits() gives n bits back to the IO system. Because 
    // n is always less than 32 this can be performed by a simple 
    // correction of the bit pointer.
    public void ungetBits(int n) {
        nextBitsAvailable = 0;
        bitPos += n;
    }

    public void findNextStartCode() throws IOException {
        if ((bitPos & 0x7) != 0) {
            bitPos &= ~0x7;
        }
        while (data != null && ! compareNextBits(24, 0x1) && ! isEOF) {
            getBits(8);
        }
    }
}
