/*
 * com/ngon/mpeg/MPEGVideoDecoder.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * This code is based on GPL'd code in MPEG_Play by Joerg Anders
 * which may be found at:
 *
 *     http://rnvs.informatik.tu-chemnitz.de/~jan/MPEG/MPEG_Play.html
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.mpeg;

import java.awt.Point;
import java.io.IOException;
import java.util.Vector;

public class MPEGVideoDecoder extends Thread {
    
    // Constants for Y, U, V channels
    final static public int Y = 0;
    final static public int U = 1;
    final static public int V = 2;
    final static public int[] CHAN_BITS = { 4, 3, 3 };
    
    // Some MPEG video layer constants
    final static protected int SEQ_END_CODE         = 0x000001b7;
    final static protected int SEQ_START_CODE       = 0x000001b3;
    final static protected int GROUP_START_CODE     = 0x000001b8;
    final static protected int FRAME_START_CODE     = 0x00000100;
    final static protected int SLICE_MIN_START_CODE = 0x00000101;
    final static protected int SLICE_MAX_START_CODE = 0x000001af;
    final static protected int EXT_START_CODE       = 0x000001b5;
    final static protected int USER_START_CODE      = 0x000001b2;
    
    final static protected int[] NULL_MATRIX = {
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0,
         0,  0,  0,  0,  0,  0,  0,  0
    };
    
    final static protected int zigzag[] = {
         0,  1,  8, 16,  9,  2,  3, 10,
        17, 24, 32, 25, 18, 11,  4,  5,
        12, 19, 26, 33, 40, 48, 41, 34,
        27, 20, 13,  6,  7, 14, 21, 28,
        35, 42, 49, 56, 57, 50, 43, 36,
        29, 22, 15, 23, 30, 37, 44, 51,
        58, 59, 52, 45, 38, 31, 39, 46,
        53, 60, 61, 54, 47, 55, 62, 63
    };
    
    final protected int defaultIntraMatrix[] = {
         8, 16, 19, 22, 26, 27, 29, 34,
        16, 16, 22, 24, 27, 29, 34, 37,
        19, 22, 26, 27, 29, 34, 34, 38,
        22, 22, 26, 27, 29, 34, 37, 40,
        22, 26, 27, 29, 32, 35, 40, 48,
        26, 27, 29, 32, 35, 40, 48, 58, 
        26, 27, 29, 34, 38, 46, 56, 69,
        27, 29, 35, 38, 46, 56, 69, 83
    };
    
    protected MPEGStream mpeg;
    
    // The pixelBuf[][][] is a main feature of the video scanner.
    // Three frames are stored at:
    //     pixelBuf[0]
    //     pixelBuf[1]
    //     pixelBuf[2]
    // The frame at index curIndex is the frame coming into being. 
    // The other two are for forward and backward prediction, and their
    // indices are kept in fwdIndex and bwdIndex respectively. 
    // The second index determines whether the information is:
    //     luminance   (Y) - pixelBuf[?][0]
    //     chrominance (U) - pixelBuf[?][1]
    //     chrominance (V) - pixelBuf[?][2]
    protected int[][][] pixelBuf;
    protected int curIndex, fwdIndex, bwdIndex;
    protected int[] pixelsPerLine;
    protected Vector framesBuf;
    protected int nrFramesRead;
    protected MPEGFrame prevIorPFrame;
    protected MPEGFrameStats frameStats;
    protected int[] huff;
    protected int[] dctRecon;
    protected int[] nonIntraQuantMatrix; 
    protected IDCT idct;
    protected MPEGMotionVector fwdMV;
    protected MPEGMotionVector bwdMV;
    protected boolean foundLumBlock;
    protected int pastDCTDCCr;
    protected int pastDCTDCCb;
    protected int pastDCTDCY;
    protected int pastDCTDVase;
    protected boolean mbDoFwdPrediction;
    protected boolean mbDoBwdPrediction;
    protected boolean mbHasCbp;
    protected boolean mbQuantFlag;
    protected boolean mbIsIntra;
    
    // Sequence parameters.
    protected int frameWidth;
    protected int frameHeight;
    protected int aspectRatio;
    protected int frameRate;
    protected int mbWidth;
    protected int mbHeight;
    protected int bitRate;
    protected int vbvBuffer; 
    protected boolean constrainedParamFlag;
    
    // Group of pictures parameters
    protected int groupTimecodeHours;
    protected int groupTimecodeMinutes;
    protected int groupTimecodeSeconds;
    protected int groupTimecodeFrameCount;
    protected boolean groupDropFrameFlag;
    protected boolean groupClosedFlag;
    protected boolean groupBrokenLinkFlag;
    
    // Frame parameters
    protected int frameType;
    protected int frameNrOffset = -1;
    protected int frameNr = 0;
    protected int vbvBufferDelay;
    
    // Slice parameters
    protected int quantScale;
    protected int mbIndex;
    protected int prevIntraMbIndex;
    
    // Macroblock parameters
    protected int mbRow;
    protected int mbCol;
    
    public MPEGVideoDecoder(MPEGStream mpeg) {
        this.mpeg = mpeg;
        
        dctRecon = new int[64];
        nonIntraQuantMatrix = new int[64]; 
        for (int i = 0; i < 64; i++) {
            nonIntraQuantMatrix[i] = 16;
        }
        
        framesBuf = new Vector();
        fwdMV = new MPEGMotionVector();
        bwdMV = new MPEGMotionVector();
        idct = new IDCT();
        idct.normalize(defaultIntraMatrix);
        idct.normalize(nonIntraQuantMatrix);
        
        prevIorPFrame = null;
        pastDCTDVase = 1024;
        
        setPriority(Thread.MIN_PRIORITY);
        setDaemon(true);
    }
    
    public void recordFrameStats(int flags) {
        this.frameStats = new MPEGFrameStats(flags);
    }
    
    public synchronized MPEGFrame nextFrame() {
        if (framesBuf == null) {
            return null;
        }
        MPEGFrame frame = null;
        if (framesBuf.size() > 0) {
            frame = (MPEGFrame) framesBuf.firstElement();
            framesBuf.removeElementAt(0);
        }
        return frame;
    }
    
    public void run() {
        Thread thisThread = Thread.currentThread();
        try {
            mpeg.findNextStartCode();
            do {
                parseSequenceHeader();
                do {
                    parseGroupOfPictures();
                } while (this == thisThread && 
                         framesBuf != null && 
                         ! mpeg.isEndOfFile() && 
                         mpeg.compareNextBits(32, GROUP_START_CODE));
            } while (this == thisThread && 
                     framesBuf != null && 
                     ! mpeg.isEndOfFile() && 
                     mpeg.compareNextBits(32, SEQ_START_CODE));
        } catch (Exception e) {
            System.err.println("MPEGVideoDecoder.run() : " + e);
            disposeFrames();
        }
        // Leftover frame.
        if (framesBuf != null && prevIorPFrame != null) {
            framesBuf.addElement(prevIorPFrame);
        }
    }
    
    public void disposeFrames() {
        if (framesBuf != null) {
            framesBuf.removeAllElements();
            framesBuf = null;
        }
    }
    
    protected void parseSequenceHeader() throws IOException {
        if (mpeg.getBits(32) != SEQ_START_CODE) {
            System.err.println("SEQ_START_CODE expected");
            return;
        }
        frameWidth           = mpeg.getBits(12);
        frameHeight          = mpeg.getBits(12);
        aspectRatio          = mpeg.getBits(4);
        frameRate            = mpeg.getBits(4);
        bitRate              = mpeg.getBits(18); 
                               mpeg.getBits(1);
        vbvBuffer            = mpeg.getBits(10);
        constrainedParamFlag = mpeg.getBits(1) == 1;
        
        mbWidth  = getMacroblockDimension(frameWidth);
        mbHeight = getMacroblockDimension(frameHeight);
        
        if (pixelBuf == null) {
            // One time initialization now we know size of vid.
            pixelBuf = new int[3][3][mbWidth * 16 * mbHeight * 16]; 
            // Can we change above to 8 x 8 for chr comps?
            pixelsPerLine = new int[3];
            pixelsPerLine[Y] = mbWidth << 4;
            pixelsPerLine[U] = mbWidth << 3;
            pixelsPerLine[V] = mbWidth << 3;
        }
        
        if (frameStats != null &&
            frameStats.isRecordingStat(MPEGFrameStats.DIMENSIONS)) {
            frameStats.setDimensions(frameWidth, frameHeight);
        }
        
        boolean intraQuantMatrixFlag = mpeg.getBits(1) == 1;
        if (intraQuantMatrixFlag) {
            for (int i = 0; i < 64; i++) {
                defaultIntraMatrix[i] = (0xff & mpeg.getBits(8));
            }
            idct.normalize(defaultIntraMatrix);
        }
        
        boolean nonIntraQuantMatrixFlag = mpeg.getBits(1) == 1;
        if (nonIntraQuantMatrixFlag) {
            for (int i = 0; i < 64; i++) {
                nonIntraQuantMatrix[i] = (0xff & mpeg.getBits(8));
            }
            idct.normalize(nonIntraQuantMatrix);
        }
        
        mpeg.findNextStartCode();
        if (mpeg.compareNextBits(32, USER_START_CODE)) {
            mpeg.getBits(32);
            while (! mpeg.compareNextBits(24, 0x1)) {
                mpeg.getBits(8);
            }
        }
    }

    protected void parseGroupOfPictures() throws IOException {
        if (mpeg.getBits(32) != GROUP_START_CODE) {
            System.err.println("GROUP_START_CODE expected");
            return;
        }
        groupDropFrameFlag      = mpeg.getBits(1) == 1;
        groupTimecodeHours      = mpeg.getBits(5);
        groupTimecodeMinutes    = mpeg.getBits(6); 
                                  mpeg.getBits(1);
        groupTimecodeSeconds    = mpeg.getBits(6);
        groupTimecodeFrameCount = mpeg.getBits(6);
        groupClosedFlag         = mpeg.getBits(1) == 1;
        groupBrokenLinkFlag     = mpeg.getBits(1) == 1;
        
        mpeg.findNextStartCode();
        if (mpeg.compareNextBits(32, EXT_START_CODE)) {
            while (! mpeg.compareNextBits(24, 0x1)) {
                mpeg.getBits(8);
            }
            System.err.println("Cannot deal with MPEG-2 data stream!");
            return;
        }
        if (mpeg.compareNextBits(32, USER_START_CODE)) {
            while (! mpeg.compareNextBits(24, 0x1)) {
                mpeg.getBits(8);
            }
        }
        
        frameNrOffset = frameNr + 2; // don't know why ???
        do {
            parseFrame();
            // An OutOfMemoryError can occur if we let
            // the number of frames grow unbounded.
            while (framesBuf != null && framesBuf.size() > 3) {
                try {
                    Thread.sleep(50);
                } catch (InterruptedException e) {}
            }
        } while (framesBuf != null && 
                 ! mpeg.isEndOfFile() && 
                 mpeg.compareNextBits(32, FRAME_START_CODE));
    }

    // The method parseFrame() parses a frame according 
    // to ISO 11172-2. It determines the frame number in display 
    // order and the frame type. Depending on the frame type 
    // some special actions are performed. Especially the index of 
    // the referred frame forward and backward prediction is to 
    // be defined.
    protected void parseFrame() throws IOException {
        int n;
        
        if (mpeg.getBits(32) != FRAME_START_CODE) {
            System.err.println("FRAME_START_CODE expected");
            return;
        }
        n = mpeg.getBits(10);
        frameNr = frameNrOffset + n;
        frameType = mpeg.getBits(3);
        vbvBufferDelay = mpeg.getBits(16);
        
        if (frameType == MPEGFrame.TYPE_P || frameType == MPEGFrame.TYPE_I) {
            fwdIndex = bwdIndex;
        }
        if (frameType == MPEGFrame.TYPE_P || frameType == MPEGFrame.TYPE_B) {
            if (fwdIndex == -1) {
                System.err.println("Warning: No predictive frame in P_FRAME");
                fwdIndex = (curIndex + 2) % 3;
            }
            fwdMV.setMotionParams(mpeg.getBits(1) == 1, mpeg.getBits(3) - 1);
        }
        if (frameType == MPEGFrame.TYPE_B) {
            if (bwdIndex == -1) {
                System.err.println("Warning: No backward predictive frame" +
                                   " in MPEGFrame.TYPE_B");
                bwdIndex = (curIndex + 1) % 3;
            }
            bwdMV.setMotionParams(mpeg.getBits(1) == 1, mpeg.getBits(3) - 1);
        }
        while (mpeg.compareNextBits(3, 0x1)) {
            mpeg.getBits(8);
        }
        mpeg.getBits(1);
        mpeg.findNextStartCode();
        
        if (mpeg.compareNextBits(32, EXT_START_CODE)) {
            mpeg.getBits(32);
            while (! mpeg.compareNextBits(24, 0x1)) {
                mpeg.getBits(8);
            }
        }
        
        if (mpeg.compareNextBits(32, USER_START_CODE)) {
            mpeg.getBits(32);
            while (! mpeg.compareNextBits(24, 0x1)) {
                mpeg.getBits(8);
            }
        }
        
        if (frameType == 4) {
            System.err.println("Can't decode D-frames!");
            return;
        }
        
        do {
            parseSlice();
            n = mpeg.getBits(32);
            mpeg.ungetBits(32);
        } while (n >= SLICE_MIN_START_CODE && 
                 n <= SLICE_MAX_START_CODE);
        
        // A frame is ready.
        MPEGFrame newFrame = new MPEGFrame(pixelBuf[curIndex], 
                                           frameWidth, frameHeight);
        // If recording frame statistics, store with frame.
        if (frameStats != null) {
            if ((frameStats.flags & MPEGFrameStats.FRAME_TYPE) != 0) {
                frameStats.setFrameType(frameType);
            }
            newFrame.setStats(frameStats.copyStats());
            frameStats.endFrame();
        }
        addFrame(newFrame);
        
        // Reorder indices.
        if (frameType == MPEGFrame.TYPE_P || frameType == MPEGFrame.TYPE_I) {
            bwdIndex = curIndex;
            curIndex = (curIndex + 1) % 3;
        }
    }

    // The method parseSlice() parses a slice according 
    // to ISO 11172-2. It determines quantization scale and 
    // the macroblock address of the first macro block of 
    // the slice.
    protected void parseSlice() throws IOException {
        int sliceStart = mpeg.getBits(32); 
        int mbNr = 0;

        // Initialization
        prevIntraMbIndex = -2; 
        pastDCTDCY  = pastDCTDVase;
        pastDCTDCCb = pastDCTDVase;
        pastDCTDCCr = pastDCTDVase;
        fwdMV.clearPreviousMotion(); 
        bwdMV.clearPreviousMotion();
        
        mbIndex = ((sliceStart & 0xff) - 1) * mbWidth - 1;
        
        if (sliceStart < SLICE_MIN_START_CODE || 
            sliceStart > SLICE_MAX_START_CODE) {
            System.err.println("SLICE START CODE expected");
            return;
        }
        quantScale = mpeg.getBits(5);
        while (mpeg.compareNextBits(1, 0x1)) {
            mpeg.getBits(1);
            mpeg.getBits(8);
        }
        mpeg.getBits(1);
        do {
            parseMacroblock(mbNr++);
        } while (! mpeg.compareNextBits(23, 0x0));
        
        mpeg.findNextStartCode();
    }
    
    // The method parseMacroblock() parses a macroblock according to 
    // ISO 11172-2. It is one of the more complicated methods because 
    // of the great variety of the constitution of a macroblock. The 
    // constitution and existence of most information fields depends on 
    // the constitution and existence of previous information fields. 
    // Furthermore the decoding process is controlled by this method. 
    // In some situations some variables must be reset to some default 
    // values or in case of skipped macroblocks implicit values must be 
    // applied. 
    // Bear in mind that some variables used in this method are member 
    // (class) variables for later reference! 
    
    // Parses off macroblock. Reconstructs DCT values. Applies
    // inverse DCT, reconstructs motion vectors, calculates and
    // set pixel values for macroblock in current pict image
    // structure.
    protected void parseMacroblock(int mbNr) throws IOException {
        int incr = 0;   // If the macro block index increment 
                        // is greater than 1 some blocks are
                        // skipped; this requires special treatment.
        int cbp = 0x3f; // Coded Block Pattern: determines which of
                        // the 6 blocks are really coded.

        while (mpeg.compareNextBits(11, 0xf)) {
            mpeg.getBits(11);
        }
        while (mpeg.compareNextBits(11, 0x8)) {
            mpeg.getBits(11);
            incr += 33;
        }
        huff = MPEGHuffman.decode(mpeg, 11, MPEGHuffman.MACROBLOCK_ADDR_INCR); 
        incr += huff[2];
        
        if (incr > 1) { // Deal with skipped macroblocks.
            // Defaults
            pastDCTDCY  = pastDCTDVase;
            pastDCTDCCr = pastDCTDVase;
            pastDCTDCCb = pastDCTDVase;
            
            if (frameType == MPEGFrame.TYPE_B && mbNr > 0) {
                // In this case the motion vectors remain valid.
                for (int off = incr - 1, idx = mbIndex + 1; 
                     off > 0; 
                     off--, idx++) {
                    int row = idx / mbWidth;
                    int col = idx % mbWidth;
                    if (mbDoFwdPrediction && ! mbDoBwdPrediction) {
                        deltPrediction(pixelBuf[fwdIndex], row, col, fwdMV, 
                                       pixelBuf[curIndex], row, col, mbWidth);
                    } else if (mbDoBwdPrediction && ! mbDoFwdPrediction) { 
                        deltPrediction(pixelBuf[bwdIndex], row, col, bwdMV, 
                                       pixelBuf[curIndex], row, col, mbWidth);
                    } else if (mbDoBwdPrediction && mbDoFwdPrediction) {
                        avgPredictions(pixelBuf[fwdIndex], fwdMV, 
                                       pixelBuf[bwdIndex], bwdMV, 
                                       pixelBuf[curIndex], row, col, mbWidth);
                    }
                }
            } else if (frameType != MPEGFrame.TYPE_I) {
                fwdMV.clearPreviousMotion();
                if (mbNr > 0 && ! mbDoBwdPrediction) {
                    for (int off = incr - 1, idx = mbIndex + 1; 
                         off > 0; 
                         off--, idx++) {
                        int row = idx / mbWidth;
                        int col = idx % mbWidth;
                        copyPrediction(pixelBuf[fwdIndex], pixelBuf[curIndex], 
                                       row, col);
                    }
                }
            }
        } // end skipped macroblocks
        
        mbIndex += incr;
        mbRow = mbIndex / mbWidth;
        mbCol = mbIndex % mbWidth;
        
        switch (frameType) {
        case MPEGFrame.TYPE_I: 
            mbDoFwdPrediction = false;
            mbDoBwdPrediction = false;
            mbHasCbp = false;
            mbIsIntra = true;
            if (mpeg.getBits(1) == 1) {
                mbQuantFlag = false;
            }  else {
                mbQuantFlag = true;
                mpeg.getBits(1);
            }
            break;
        case MPEGFrame.TYPE_P: 
            huff = MPEGHuffman.decode(mpeg, 6, MPEGHuffman.MACROBLOCK_TYPE_P);
            mbQuantFlag = huff[2] != 0;
            mbDoFwdPrediction = (huff[3] == 1);
            mbDoBwdPrediction = false;
            mbHasCbp = (huff[4] == 1);
            if (! (mbIsIntra = huff[5] != 0)) {
                // Default values
                pastDCTDCY = pastDCTDCCr = pastDCTDCCb = pastDCTDVase;
                cbp = 0;
            }
            break;
        case MPEGFrame.TYPE_B: 
            huff = MPEGHuffman.decode(mpeg, 6, MPEGHuffman.MACROBLOCK_TYPE_B);
            mbQuantFlag = huff[2] != 0;
            mbDoFwdPrediction = (huff[3] == 1);
            mbDoBwdPrediction = (huff[4] == 1);
            mbHasCbp = (huff[5] == 1);
            if (!(mbIsIntra = huff[6] != 0)) {
                // Default values
                pastDCTDCY = pastDCTDCCr = pastDCTDCCb = pastDCTDVase;
                cbp = 0;
            }
            break;
        default: 
            System.err.println("Unknown frame type: " + frameType);
            return;
        }
        
        if (mbQuantFlag) {
            // Extract new quantization factor
            quantScale = mpeg.getBits(5);
        }
        
        if (mbDoFwdPrediction) {
            int horzCode, horzR, vertCode, vertR;
            huff = MPEGHuffman.decode(mpeg, 11, MPEGHuffman.MOTION_CODE);
            horzCode = huff[2];
            horzR = (fwdMV.f != 1 && horzCode != 0) ? mpeg.getBits(fwdMV.rBits) 
                                                    : 0;
            huff = MPEGHuffman.decode(mpeg, 11, MPEGHuffman.MOTION_CODE);
            vertCode = huff[2];
            vertR = (fwdMV.f != 1 && vertCode != 0) ? mpeg.getBits(fwdMV.rBits) 
                                                    : 0;
            fwdMV.decodeMotionVector(horzCode, vertCode, horzR, vertR);
            
            if (frameType != MPEGFrame.TYPE_B || ! mbDoBwdPrediction) {
                deltPrediction(pixelBuf[fwdIndex], mbRow, mbCol, fwdMV,
                               pixelBuf[curIndex], mbRow, mbCol, mbWidth);
            }
        } else if (frameType != MPEGFrame.TYPE_B) {
            fwdMV.clearPreviousMotion();
        }
        
        // Optional motion vector statistics for forward motion.
        if (frameStats != null &&
            frameStats.isRecordingStat(MPEGFrameStats.MOTION_VECTORS)) {
            Point mv = mbDoFwdPrediction ? fwdMV.getMotionVector() 
                                         : new Point(0, 0);
            frameStats.setMotionVector(mbRow, mbCol, mv);
        }
        
        if (mbDoBwdPrediction) {
            int horzCode, horzR, vertCode, vertR;
            huff = MPEGHuffman.decode(mpeg, 11, MPEGHuffman.MOTION_CODE);
            horzCode = huff[2];
            horzR = (bwdMV.f != 1 && horzCode != 0) ? mpeg.getBits(bwdMV.rBits) 
                                                    : 0;
            huff = MPEGHuffman.decode(mpeg, 11, MPEGHuffman.MOTION_CODE);
            vertCode = huff[2];
            vertR = (bwdMV.f != 1 && vertCode != 0) ? mpeg.getBits(bwdMV.rBits) 
                                                    : 0;
            bwdMV.decodeMotionVector(horzCode, vertCode, horzR, vertR);
            
            if (! mbDoFwdPrediction) {
                deltPrediction(pixelBuf[bwdIndex], mbRow, mbCol, bwdMV, 
                               pixelBuf[curIndex], mbRow, mbCol, mbWidth);
            } else {
                avgPredictions(pixelBuf[fwdIndex], fwdMV, 
                               pixelBuf[bwdIndex], bwdMV, 
                               pixelBuf[curIndex], mbRow, mbCol, mbWidth);
            }
        }
        
        if (mbHasCbp) {
            huff = MPEGHuffman.decode(mpeg, 9, MPEGHuffman.CODED_BLOCK_PATTERN);
            cbp = huff[2];
        }

        if (frameType == MPEGFrame.TYPE_P && 
            ! mbDoBwdPrediction && 
            ! mbDoFwdPrediction) {
            copyPrediction(pixelBuf[fwdIndex], pixelBuf[curIndex],
                           mbRow, mbCol);
        }
        foundLumBlock = false;
        for (int i = 0; i < 6; i++) {
            // Block information supplied?
            if ((cbp & (1 << (5 - i))) != 0) {
                // Yes => get block information.
                parseBlock(i);
                if (mbIsIntra) {
                    // In intra-coded macro blocks the values are absolute.
                    if (i < 4) {
                        setLumBlock(i);
                    } else {
                        setChrBlock(i);
                    }
                } else { 
                    // In inter-coded macro blocks the values are deltas.
                    if (i < 4) {
                        deltLumBlock(i);
                    } else {
                        deltChrBlock(i);
                    }
                }
            }
        }
        if (frameType == MPEGFrame.TYPE_B && mbIsIntra) {
            fwdMV.clearPreviousMotion(); 
            bwdMV.clearPreviousMotion();
        }
    }
    
    // The method parseBlock() parses a block according to ISO 11172-2. 
    // Thereby the DC and AC coefficients are reconstructed and placed 
    // into the dctRecon[] array in de-"zigzag"-ed order. After that 
    // the IDCT routine is called. The method counts the coefficients 
    // and calls a sparse IDCT method if the coefficient count is 1. 
    protected void parseBlock(int nr) throws IOException {
        int sign;
        int idx = 0;
        int level;
        int coeffCount = 0;
        int pos = 0;
        
        // Initialization
        System.arraycopy(NULL_MATRIX, 0, dctRecon, 0, 64);
        
        if (mbIsIntra) {
            int dcSize;
            if (nr < 4) { 
                // Luminance block
                huff = MPEGHuffman.decode(mpeg, 7, 
                                          MPEGHuffman.DCT_SIZE_LUMINANCE);
                dcSize = huff[2]; // dcSize of the DC coefficient
                if (dcSize != 0) {
                    setDCTDC(mpeg.getBits(dcSize), dcSize);
                }
                if (foundLumBlock) { 
                    // Not first luminance block
                    dctRecon[0] = pastDCTDCY + (dctRecon[0] << 3);
                    pastDCTDCY = dctRecon[0];
                } else {
                    // First luminance block
                    foundLumBlock = true;
                    dctRecon[0] <<= 3;
                    if (mbIndex - prevIntraMbIndex > 1) {
                        dctRecon[0] += pastDCTDVase;
                        pastDCTDCY = dctRecon[0];
                    } else { 
                        // Relative if no skipping.
                        dctRecon[0] += pastDCTDCY;
                        pastDCTDCY = dctRecon[0];
                    }
                }
                prevIntraMbIndex = mbIndex;
            } else {
                // Chrominance block
                huff = MPEGHuffman.decode(mpeg, 8, 
                                          MPEGHuffman.DCT_SIZE_CHROMINANCE);
                dcSize = huff[2];
                if (dcSize != 0) {
                    setDCTDC(mpeg.getBits(dcSize), dcSize);
                }
                switch (nr) {
                case 4: 
                    dctRecon[0] <<= 3;
                    if (mbIndex - prevIntraMbIndex > 1) {
                        pastDCTDCCb = dctRecon[0] += pastDCTDVase;
                    } else {
                        pastDCTDCCb = dctRecon[0] += pastDCTDCCb;
                    }
                    break;
                case 5: 
                    dctRecon[0] <<= 3;
                    if (mbIndex - prevIntraMbIndex > 1) {
                        pastDCTDCCr = dctRecon[0] += pastDCTDVase;
                    } else {
                        pastDCTDCCr = dctRecon[0] += pastDCTDCCr;
                    }
                    break;
                }
            }
            prevIntraMbIndex = mbIndex;
            
            if (dctRecon[0] != 0) {
                coeffCount = 1;
            }
            
            // Because of the IDCT technique,
            // the DC values are not quantized;
            // therefore the fix point translation
            // must be performed.
            dctRecon[0] <<= idct.VAL_BITS - 3;
            
        } else {
            // Not intra-coded block => first AC value.
            if (mpeg.compareNextBits(1, 0x1)) { 
                // Special treatment of the VLC "1".
                idx = 0; 
                mpeg.getBits(1);
                sign = level = mpeg.getBits(1) == 0 ? 1 : -1;
            } else {
                huff = MPEGHuffman.decode(mpeg, 28, MPEGHuffman.DCT_COEFF);
                idx = huff[2];
                if (idx == MPEGHuffman.ESC) {
                    idx = mpeg.getBits(6);
                    if (((level = mpeg.getBits(8)) & 0x7f) == 0) {
                        level <<= 8;
                        level |= mpeg.getBits(8);
                        if ((level & 0x8000) != 0) {
                            level |= 0xffffff00;
                        }
                    } else if ((0x80 & level) != 0) {
                        level |= 0xffffff00;
                    }
                } else { 
                    // Normal treatment (no escape); extract AC coefficient.
                    level = (mpeg.getBits(1) == 0) ? huff[3] : -huff[3];
                }
                sign = (level == 0) ? 0 : ((level < 0) ? -1 : 1);
            }
            pos = zigzag[idx]; // de-zigzag

            // Quantization:
            dctRecon[pos] = ((level + sign) * 
                             quantScale *
                             nonIntraQuantMatrix[pos]) >> 3;

            if (level != 0) {
                // Count the coefficients.
                coeffCount++;
            }
        }
        
        // Decode AC value.
        huff = MPEGHuffman.decode(mpeg, 28, MPEGHuffman.DCT_COEFF); 
        
        int idxRun;
        while ((idxRun = huff[2]) != MPEGHuffman.EOB) {
            if (idxRun == MPEGHuffman.ESC) {
                idxRun = mpeg.getBits(6);
                if (((level = mpeg.getBits(8)) & 0x7f) == 0) {
                    // 16 bit
                    level <<= 8;
                    level |= mpeg.getBits(8);
                    if ((level & 0x8000) != 0) {
                        level |= 0xffffff00;
                    }
                } else if ((0x80 & level) != 0) {
                    level |= 0xffffff00;
                }
                idx += idxRun + 1;
            } else { 
                // Normal treatment (no escape); extract AC coefficient.
                idx += idxRun + 1;
                level = (mpeg.getBits(1) == 0) ? huff[3] : -huff[3]; 
            }
            if (idx > 63) {
                idx = 63;
            }
            pos = zigzag[idx];
            if (mbIsIntra) {
                // Different treatment of quantization depending on type.
                dctRecon[pos] = 
                    (level * quantScale * defaultIntraMatrix[pos]) >> 3;
            } else {
                sign = (level == 0) ? 0 : ((level < 0) ? -1 : 1); 
                dctRecon[pos] = 
                    ((level + sign) * quantScale * nonIntraQuantMatrix[pos])
                        >> 3;
            }

            if (level != 0) {
                coeffCount++;
            }
            huff = MPEGHuffman.decode(mpeg, 28, MPEGHuffman.DCT_COEFF);
        }
        if (coeffCount == 1) { 
            // Call a sparse method.
            idct.inverseDCTSparse(dctRecon, pos); 
        } else {
            idct.inverseDCT(dctRecon);
        }
    }
    
    private synchronized void addFrame(MPEGFrame newFrame) {
        if (framesBuf == null) {
            return;
        }
        if (framesBuf.size() == 0) {
            framesBuf.addElement(newFrame);
            nrFramesRead = 1;
        } else if (frameType == MPEGFrame.TYPE_B) {
            framesBuf.addElement(newFrame);
            nrFramesRead++;
        } else {
            if (prevIorPFrame != null) {
                framesBuf.addElement(prevIorPFrame);
                nrFramesRead++;
            }
            prevIorPFrame = newFrame;
        }
    }
    
    private void deltPrediction(int[][] srcMB, int srcRow, int srcCol, 
                                MPEGMotionVector mv, int[][] dstMB, 
                                int dstRow, int dstCol, int dstMBWidth) {
        for (int ch = 0; ch < 3; ch++) {
            int[] src = srcMB[ch];
            int srcX = (srcCol << CHAN_BITS[ch]) + mv.motionRight[ch];
            int srcY = (srcRow << CHAN_BITS[ch]) + mv.motionDown[ch];
            int srcScan = pixelsPerLine[ch];
            int srcPos = srcScan * srcY + srcX;
            int[] dst = dstMB[ch];
            int dstScan = dstMBWidth << CHAN_BITS[ch];
            int dstPos = dstScan * (dstRow << CHAN_BITS[ch]) + 
                         (dstCol << CHAN_BITS[ch]);
            int blockSize = 1 << CHAN_BITS[ch];
            
            if (! mv.doHalfStepRight[ch] && ! mv.doHalfStepDown[ch]) {
                // No half blockSize; straight copy
                for (int j = 0; j < blockSize; j++) {
                    System.arraycopy(src, srcPos, dst, dstPos, blockSize);
                    srcPos += srcScan;
                    dstPos += dstScan;
                }
            } else if (! mv.doHalfStepRight[ch] && 
                       mv.doHalfStepDown[ch])  {
                int srcPosDown = srcPos + srcScan;
                for (int j = 0; j < blockSize; j++) {
                    for (int i = 0; i < blockSize; i++) {
                        dst[dstPos++] = (src[srcPos++] + 
                                         src[srcPosDown++]) >> 1;
                    }
                    srcPos     += srcScan - blockSize; 
                    srcPosDown += srcScan - blockSize; 
                    dstPos     += dstScan - blockSize;
                }
            } else if (mv.doHalfStepRight[ch] && 
                       ! mv.doHalfStepDown[ch]) {
                int srcPosRight = srcPos + 1;
                for (int j = 0; j < blockSize; j++) {
                    for (int i = 0; i < blockSize; i++) {
                        dst[dstPos++] = (src[srcPos++] + 
                                         src[srcPosRight++]) >> 1;
                    }
                    srcPos      += srcScan - blockSize; 
                    srcPosRight += srcScan - blockSize; 
                    dstPos      += dstScan - blockSize;
                }
            } else {
                int srcPosDown      = srcPos + srcScan;
                int srcPosRight     = srcPos + 1;
                int srcPosDownRight = srcPosDown + 1;
                for (int j = 0; j < blockSize; j++) {
                    for (int i = 0; i < blockSize; i++) {
                        dst[dstPos++] = (src[srcPos++] + 
                                         src[srcPosDown++] +
                                         src[srcPosRight++] + 
                                         src[srcPosDownRight++]) >> 2;
                    }
                    srcPos          += srcScan - blockSize; 
                    srcPosDown      += srcScan - blockSize; 
                    srcPosRight     += srcScan - blockSize; 
                    srcPosDownRight += srcScan - blockSize;
                    dstPos          += dstScan - blockSize;
                }
            }
        }
    }
    
    private void avgPredictions(int[][] srcMB1, MPEGMotionVector mv1,
                                int[][] srcMB2, MPEGMotionVector mv2,
                                int[][] dstMB, int row, int col, int mbWidth) {
        // Fill temporary buffers with ref predictions.
        int[][] tmpMB1 = new int[3][];
        int[][] tmpMB2 = new int[3][];
        for (int ch = 0; ch < 3; ch++) {
            int blockSize = 1 << CHAN_BITS[ch];
            tmpMB1[ch] = new int[blockSize * blockSize];
            tmpMB2[ch] = new int[blockSize * blockSize];
        }
        deltPrediction(srcMB1, row, col, mv1, tmpMB1, 0, 0, 1);
        deltPrediction(srcMB2, row, col, mv2, tmpMB2, 0, 0, 1);
        
        // Average and store in dstMB.
        for (int ch = 0; ch < 3; ch++) {
            int srcPos = 0;
            int dstScan = mbWidth << CHAN_BITS[ch];
            int dstPos = dstScan * (row << CHAN_BITS[ch]) + 
                         (col << CHAN_BITS[ch]);
            int blockSize = 1 << CHAN_BITS[ch];

            for (int j = 0; j < blockSize; j++) {
                for (int i = 0; i < blockSize; i++) {
                    dstMB[ch][dstPos++] = (tmpMB1[ch][srcPos] + 
                                           tmpMB2[ch][srcPos]) >> 1;
                    srcPos++;
                }
                dstPos += dstScan - blockSize;
            }
        }
    }
    
    // The method copyPrediction() grabs from the srcRef and puts macroblock  
    // in dstMB.
    private void copyPrediction(int[][] srcMB, int[][] dstMB, 
                                int row, int col) {
        for (int ch = 0; ch < 3; ch++) {
            int[] src = srcMB[ch];
            int[] dst = dstMB[ch];
            int p = pixelsPerLine[ch] * (row << CHAN_BITS[ch]) + 
                    (col << CHAN_BITS[ch]);
            int blockSize = 1 << CHAN_BITS[ch];
            for (int j = 0; j < blockSize; j++) {
                System.arraycopy(src, p, dst, p, blockSize);
                p += pixelsPerLine[ch];
            }
        }
    }
    
    // The method setDCTDC() computes the DCT DC according to 
    // ISO 11172-2. That is, it sets dctRecon[0].
    private void setDCTDC(int dctDiff, int dctSize) {
        if ((dctDiff & (1 << (dctSize - 1))) != 0) {
            dctRecon[0] = dctDiff;
        } else {
            dctRecon[0] = (-1 << (dctSize)) | (dctDiff + 1);
        }
    }

    // The method setLumBlock() takes the re-transformed luminance values 
    // and places them at the appropriate position.
    private void setLumBlock(int nr) {
        if (mbRow >= mbHeight) {
            // Must be a bug somewhere.  This sometimes occurs!
            System.err.println("MPEGVideoDecoder.setLumBlock() : " +
                "mbRow >= mbHeight : " +
                "nr=" + nr +
                " mbRow=" + mbRow + " mbCol=" + mbCol +
                " mbIndex=" + mbIndex +
                " curIndex=" + curIndex);
            return;
        }
        
        int p;
        int q = pixelsPerLine[Y] * ((mbRow << 4) + ((nr & 0x2) << 2)) +
                (mbCol << 4) + ((nr & 0x1) << 3);
        for (p = 0; p < 64; p += 8) {
            System.arraycopy(dctRecon, p, pixelBuf[curIndex][Y], q, 8);
            q += pixelsPerLine[Y];
        }
    }

    // The method setChrBlock() takes the re-transformed chrominance values 
    // and places them at the appropriate position.
    private void setChrBlock(int nr) {
        if (mbRow >= mbHeight) {
            // See note in setLumBlock().
            return;
        }
        int p;
        int q = pixelsPerLine[U] * (mbRow << 3) + (mbCol << 3);
        switch (nr) {
        case 4: 
            for (p = 0; p < 64; p += 8) {
                System.arraycopy(dctRecon, p, pixelBuf[curIndex][V], q, 8);
                q += pixelsPerLine[V];
            }
            break;
        case 5: 
            for (p = 0; p < 64; p += 8) {
                System.arraycopy(dctRecon, p, pixelBuf[curIndex][U], q, 8);
                q += pixelsPerLine[U];
            }
            break;
        }
    }
        
    // The method deltLumBlock() is called in predicted macro blocks. 
    // Because the values in dctRecon are motion compensation information the 
    // task of this method is to delta the already supplied (copied) values.
    private void deltLumBlock(int nr) {
        if (mbRow >= mbHeight) {
            // See note in setLumBlock().
            return;
        }
        int i, j, p = 0;
        int q = pixelsPerLine[Y] * ((mbRow << 4) + ((nr & 0x2) << 2)) +
                (mbCol << 4) + ((nr & 0x1) << 3);

        for (j = 0; j < 8; j++) {
            for (i = 0; i < 8; i++) {
                pixelBuf[curIndex][0][q++] += dctRecon[p++];
            }
            q += pixelsPerLine[Y] - 8;
        }
    }

    // The method deltChrBlock() is called in predicted macro blocks. 
    // Because the values in dctRecon are motion compensation information the 
    // task of this method is to delta the already supplied (copied) values.
    private void deltChrBlock(int nr) {
        if (mbRow >= mbHeight) {
            // See note in setLumBlock().
            return;
        }
        int i, j, p = 0;
        int q = pixelsPerLine[U] * (mbRow << 3) + (mbCol << 3);
        
        switch (nr) {
        case 4: 
            for (j = 0; j < 8; j++) {
                for (i = 0; i < 8; i++) {
                    pixelBuf[curIndex][V][q++] += dctRecon[p++];
                }
                q += pixelsPerLine[V] - 8;
            }
            break;
        case 5: 
            for (j = 0; j < 8; j++) {
                for (i = 0; i < 8; i++) {
                    pixelBuf[curIndex][U][q++] += dctRecon[p++];
                }
                q += pixelsPerLine[U] - 8;
            }
            break;
        }
    }
    
    static public int getMacroblockDimension(int d) {
        return (d + 15) >> 4;
    }
    
    static public int getMacroblockSize() {
        return 16;
    }
}
