/*
 * com/ngon/tap/ClipSeg.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.app.App;
import com.ngon.mpeg.MPEGFrame;
import com.ngon.mpeg.MPEGDecoder;
import java.io.InputStream;

/**
 * class ClipSeg
 */
class ClipSeg extends Clip {
    
    final static public int N_CLIPSEG_FRAMES    = 10;
    final static public int PRELOAD_BUFFER      = 3;
    final static public int N_CLIPSEG_NR_DIGITS = 4;
    
    // Frame # for current clip segment [1, N_CLIPSEG_FRAMES].
    protected int segFrameNr;
    
    protected InputStream[] mpegsOnDeck;
    protected MPEGDecoder[] decsOnDeck;
    protected boolean onDeck;
    
    ClipSeg(String clipsPath, String clipName, App app) {
        super(clipsPath, clipName, app);
    }
    
    public void init() {
        super.init();
        // Also preload next if close to end.
        if (segFrameNr > N_CLIPSEG_FRAMES-2) {
            preloadNextClipSeg();
        }
    }
    
    public void setFrameNr(int frameNr) {
        super.setFrameNr(frameNr);
        this.segFrameNr = frameNr % N_CLIPSEG_FRAMES;
    }
    
    public void skipFrames(int n) {
        super.skipFrames(n % N_CLIPSEG_FRAMES);
    }
    
    public boolean nextFrame() {
        // If near the end of the clip segment, go to next.
        if (onDeck && segFrameNr == N_CLIPSEG_FRAMES-1) {
            // Kill the current stream.
            decs[0].killVideoDecoder();
            decs[0] = null;
            try {
                mpegs[0].close();
            } catch (Exception e) {
                System.err.println("Exception closing mpeg : " + e);
            }
            
            mpegs = mpegsOnDeck;
            decs = decsOnDeck;
            mpegsOnDeck = null;
            decsOnDeck = null;
            onDeck = false;
        }
        boolean success = super.advanceDecs(nRetries);
        if (success) {
            incrementFrameNr(); 
            // Begin preload of next segment if close to end of current.
            if (! onDeck && segFrameNr > N_CLIPSEG_FRAMES - PRELOAD_BUFFER) {
                preloadNextClipSeg();
            }
        }
        return success;
    }
    
    public String[] getMPEGFilenames(String clipsPath, String clipName, 
                                     int frameNr) {
        String[] mpegFilenames = new String[nStreams];
        int clipSegNr = (frameNr / N_CLIPSEG_FRAMES) + 1;
        StringBuffer clipFileAppend = new StringBuffer();
        clipFileAppend.append(clipSegNr);
        while (clipFileAppend.length() < N_CLIPSEG_NR_DIGITS) {
            clipFileAppend.insert(0, '0');
        }
        mpegFilenames[0] = new String(
            clipsPath + clipName + "/" + N_CLIPSEG_FRAMES + "/" +
            clipName + "-" + clipFileAppend + ".mpg");
        return mpegFilenames;
    }
    
    /**
     * mpeg_encode won't create a clip segment with one frame, so for 
     * safety, add a buffer.
     */
    public boolean isLastFrame() {
        return (this.frameNr >= this.clipLength-2);
    }
    
    public void kill() {
        for (int i = 0; i < nStreams; i++) {
            if (mpegsOnDeck != null && mpegsOnDeck[i] != null) {
                try {
                    mpegsOnDeck[i].close();
                    mpegsOnDeck[i] = null;
                } catch (Exception e) {
                    System.err.println("Exception closing mpeg : " + e);
                }
            }
            if (decsOnDeck != null && decsOnDeck[i] != null) {
                decsOnDeck[i].killVideoDecoder();
                decsOnDeck[i] = null;
            }
        }
        mpegsOnDeck = null;
        decsOnDeck = null;
        super.kill();
    }
    
    protected void incrementFrameNr() {
        super.incrementFrameNr();
        segFrameNr = frameNr % N_CLIPSEG_FRAMES;
    }
    
    protected void preloadNextClipSeg() {
        int nextClipSegStart =
            ((frameNr / N_CLIPSEG_FRAMES) + 1) * N_CLIPSEG_FRAMES;
        if (Debug.DEBUG) {
            app.println("preloadNextClipSeg called with frameNr=" + frameNr +
                " nextClipSegStart=" + nextClipSegStart + 
                " clipLength=" + clipLength);
        }
        if (Debug.DEBUG_APPLET) {
            System.err.println("preloadNextClipSeg called with frameNr=" + frameNr +
                " nextClipSegStart=" + nextClipSegStart + 
                " clipLength=" + clipLength);
        }
        if (nextClipSegStart > clipLength-2) {
            return;
        }
        mpegsOnDeck = loadMPEGs(clipsPath, clipName, nextClipSegStart);
        if (mpegsOnDeck == null) {
            return;
        }
        decsOnDeck = createDecs(mpegsOnDeck);
        for (int i = 0; i < nStreams; i++) {
            decsOnDeck[i].start();
        }
        onDeck = true;
    }
}

