/*
 * com/ngon/tap/ClusQuery.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.math.Math1D;
import com.ngon.math.Math2D;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Vector;

/**
 * class ClusQuery
 */
abstract class ClusQuery {
    
    final static String MVCLUSTER_FIELDS = 
        "id,n_blocks,width,height,cg_x,cg_y,avg_mv_x,avg_mv_y," +
        "avg_div,avg_curl,avg_rgb,frame_nr,clip_id,link_wt_sum";
    final static String MVCLUSTER_QUERY = 
        "SELECT " + MVCLUSTER_FIELDS + " FROM mvcluster";
    final static String MVCLUSTER_QUERY_BY_ID = 
        MVCLUSTER_QUERY + " WHERE id = ?";
    
    protected Connection dbc = null;
    protected String query = null;
    protected PreparedStatement preparedStmt = null;
    protected int[] bindVars = null;
    protected Results results = null;
    protected boolean isVerbose = false;
    
    ClusQuery(Connection dbc) {
        this.dbc = dbc;
        this.results = new Results();
    }
    
    abstract void fillBindVars(Clus ref);
    
    Clus selectCluster(Clus ref) {
        findResults(ref);
        if (results == null) {
            return null;
        }
        if (isVerbose) {
            System.out.println("# possible links = " + results.size());
        }
        return results.getChoice();
    }
    
    Vector findResults(Clus ref) {
        if (preparedStmt == null || bindVars == null) {
            init();
        }
        fillBindVars(ref);
        results.setReference(ref);
        ResultSet rs = null;
        try {
            setBindVars();
            rs = preparedStmt.executeQuery();
            while (rs.next()) {
                results.addElement(Clus.parseRow(rs));
            }
            rs.close();
            rs = null;
        } catch (Exception e) {
            System.err.println(e);
            return null;
        } finally {
            if (rs != null) {
                try { rs.close(); } catch (SQLException e) {}
                rs = null;
            }
        }
        return getResults();
    }
    
    Vector getResults() {
        return (Vector) results;
    }
    
    int getResultsWeightTotal() {
        return results.weightTotal;
    }
    
    void setVerbose(boolean b) {
        this.isVerbose = b;
    }
    
    void close() {
        try {
            if (preparedStmt != null) {
                preparedStmt.close();
                preparedStmt = null;
            }
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (preparedStmt != null) {
                try { preparedStmt.close(); } catch (SQLException e) {}
                preparedStmt = null;
            }
        }
        bindVars = null;
        if (results != null) {
            results.removeAllElements();
            results = null;
        }
    }
    
    protected void init() {
        try {
            this.preparedStmt = dbc.prepareStatement(query);
        } catch (SQLException e) {
//            System.err.println(e);
            e.printStackTrace();
        }
        
        // Count # of ?'s in query to determine # of bind vars.
        int nrBinds = 0;
        int p = -1;
        while ((p = query.indexOf('?', p+1)) != -1) {
            nrBinds++;
        }
        this.bindVars = new int[nrBinds];
    }
    
    protected void setBindVars() throws SQLException {
        for (int i = 0; i < bindVars.length; i++) {
            preparedStmt.setInt(i+1, bindVars[i]);
        }
    }
    
    final private class Results extends Vector {
        int weightTotal;
        Clus ref;
        
        void setReference(Clus ref) {
            removeAllElements();
            this.weightTotal = 0;
            this.ref = ref;
            if (isVerbose) {
                System.out.println(
                    "========================================" + 
                    "========================================");
                System.out.println("Reference: clip " + ref.clipID + 
                                   " frame " + ref.frameNr);
                System.out.println(" cg=("+ref.cg.x+","+ref.cg.y+")"+
                                   " ["+ref.blockSize+"] "+
                                   ref.width+"x"+ref.height+
                                   " mv=("+ref.avgMV.x+","+ref.avgMV.y+")"+
                                   " div="+ref.avgDiv+
                                   " curl="+ref.avgCurl);
                System.out.println(
                    "----------------------------------------" + 
                    "----------------------------------------");
            }
        }
        
        public void addElement(Object o) {
            Clus c = (Clus) o;
            c.weight = determineProbabilityWeight(c);
            if (c.weight > 0) {
                weightTotal += c.weight;
                super.addElement(o);
            }
        }
        
        int determineProbabilityWeight(Clus c) {
            int blockSizeFactor = (c.blockSize + 95) / 100;
            int widthFactor = Math.max(3 - Math.abs(c.width - ref.width), 0);
            int heightFactor = Math.max(3 - Math.abs(c.height - ref.height), 0);
            
            // Base weight
            int wt0 = 0;
            
            // cg contribution
            int wtCG = 0;
            int d2 = Math1D.sqr(c.cg.x - ref.cg.x) + 
                     Math1D.sqr(c.cg.y - ref.cg.y);
            int m = Math.max(15488 - d2, 0) >> 12;
            wtCG = 
                (int) m * m *
                blockSizeFactor *
                (widthFactor + 1) *
                (heightFactor + 1);
            
            // blockSize contribution
            int wtSz = 0;
            if (wtCG > 0) {
                wtSz = Math.max(115 - Math.abs(c.blockSize - ref.blockSize), 
                                0) *
                       (blockSizeFactor-1);
            }
            
            // width & height contributions
            int wtW = (wtCG > 0 && c.width > 1) ? widthFactor * 16 : 0;
            int wtH = (wtCG > 0 && c.height > 1) ? heightFactor * 16 : 0;
            
            // mv contribution
            int wtMV = 0;
            if (Math.abs(Math2D.angleBetween(c.avgMV, ref.avgMV)) < 45) {
                d2 = Math1D.sqr(c.avgMV.x - ref.avgMV.x) + 
                     Math1D.sqr(c.avgMV.y - ref.avgMV.y);
                wtMV = (int) Math.sqrt(Math.max(36 - d2, 0)) *
                       blockSizeFactor *
                       (widthFactor + 1) *
                       (heightFactor + 1) *
                       4;
            }
            
            // div contribution
            int wtDiv = 0;
            if (wtCG > 0) {
                wtDiv = 
                    Math.max(3 - Math.abs(c.avgDiv - ref.avgDiv), 0) *
                    (Math.abs(c.avgDiv) + 1) *
                    (Math.abs(ref.avgDiv) + 1) *
                    blockSizeFactor *
                    (widthFactor + 1) *
                    (heightFactor + 1);
            }
            
            // curl contribution
            int wtCurl = 0;
            if (wtCG > 0) {
                wtCurl = 
                    Math.max(2 - Math.abs(c.avgCurl - ref.avgCurl), 0) *
                    (Math.abs(c.avgCurl) + 1) *
                    (Math.abs(ref.avgCurl) + 1) *
                    blockSizeFactor *
                    (widthFactor + 1) *
                    (heightFactor + 1);
            }
            
            if (isVerbose) {
                System.out.println("Candidate: clip " + c.clipID + 
                                   " frame " + c.frameNr);
                System.out.println(" cg=("+c.cg.x+","+c.cg.y+")"+
                                   " ["+c.blockSize+"] "+c.width+"x"+c.height+
                                   " mv=("+c.avgMV.x+","+c.avgMV.y+")"+
                                   " div="+c.avgDiv+
                                   " curl="+c.avgCurl);
                System.out.print(" ");
                printTics(wt0, '0');
                printTics(wtCG, 'G');
                printTics(wtSz, 'N');
                printTics(wtW, 'W');
                printTics(wtH, 'H');
                printTics(wtMV, 'V');
                printTics(wtDiv, 'D');
                printTics(wtCurl, 'C');
                System.out.println();
            }
            
            return wt0 + wtSz + wtW + wtH + wtCG + wtMV + wtDiv + wtCurl;
        }
        
        Clus getChoice() {
            if (size() == 0) {
                return null;
            }
            if (size() == 1) {
                return (Clus) elementAt(0);
            }
            
            int rand = (int) (Math.random() * weightTotal);
            int i = 0;
            Clus choice;
            do {
                choice = (Clus) elementAt(i++);
                rand -= choice.weight;
            } while (rand > -1);
            if (isVerbose) {
                System.out.println("Choosing clus with clipID " + 
                    choice.clipID + " and frame " + choice.frameNr);
                System.out.println(" ref was           clipID " + 
                    ref.clipID + " and frame " + ref.frameNr);
            }
            return choice;
        }
    }
    
    private void printTics(int n, char c) {
        for (int i = 0; i < n; i++) {
            System.out.print(c);
        }
    }
}

class ClusQueryInSequence extends ClusQuery {
    
    final static String QUERY = 
        MVCLUSTER_QUERY + " WHERE (clip_id = ?) AND (frame_nr = ?)";
    
    ClusQueryInSequence(Connection dbc) {
        super(dbc);
        this.query = QUERY;
    }
    
    void fillBindVars(Clus ref) {
        bindVars[0] = ref.clipID;
        bindVars[1] = ref.frameNr + 1;
    }
}

class ClusQueryOutOfSequence extends ClusQuery {
    
    final static String QUERY = 
        MVCLUSTER_QUERY + 
        " WHERE (clip_id != ?) AND " +
        "(n_blocks BETWEEN ? AND ?) AND " +
        "(avg_mv_x BETWEEN ? AND ?) AND (avg_mv_y BETWEEN ? AND ?) AND " +
        "(cg_x BETWEEN ? AND ?) AND (cg_y BETWEEN ? AND ?) AND " +
        "(width BETWEEN ? AND ?) AND (height BETWEEN ? AND ?) AND " +
        "(avg_div BETWEEN ? AND ?) AND (avg_curl BETWEEN ? AND ?)";
    
    static int BLOCK_SZ_RANGE =   4; // This is modified by a mult factor
    static int AVG_MV_RANGE   =   1;
    static int CG_RANGE       =   8;
    static int WIDTH_RANGE    =   2;
    static int HEIGHT_RANGE   =   1;
    static int AVG_DIV_RANGE  =   1; // This is modified by a mult factor
    static int AVG_CURL_RANGE =   1; // This is modified by a mult factor
    
    ClusQueryOutOfSequence(Connection dbc) {
        super(dbc);
        this.query = QUERY;
    }
    
    void fillBindVars(Clus ref) {
        int mult;
        bindVars[0] = ref.clipID;
        mult = (ref.blockSize / 20) + 1;
        bindVars[1] = ref.blockSize - BLOCK_SZ_RANGE * mult;
        bindVars[2] = ref.blockSize + BLOCK_SZ_RANGE * mult;
        bindVars[3] = ref.avgMV.x - AVG_MV_RANGE;
        bindVars[4] = ref.avgMV.x + AVG_MV_RANGE;
        bindVars[5] = ref.avgMV.y - AVG_MV_RANGE;
        bindVars[6] = ref.avgMV.y + AVG_MV_RANGE;
        bindVars[7] = ref.cg.x - CG_RANGE;
        bindVars[8] = ref.cg.x + CG_RANGE;
        bindVars[9] = ref.cg.y - CG_RANGE;
        bindVars[10] = ref.cg.y + CG_RANGE;
        bindVars[11] = ref.width - WIDTH_RANGE;
        bindVars[12] = ref.width + WIDTH_RANGE;
        bindVars[13] = ref.height - HEIGHT_RANGE;
        bindVars[14] = ref.height + HEIGHT_RANGE;
        // Div and curl ranges depend on the ref value.
        mult = ((Math.abs(ref.avgDiv)-1) / 4) + 1;
        if (ref.avgDiv == 0) {
            bindVars[15] = -1;
            bindVars[16] = 1;
        } else {
            bindVars[15] = ref.avgDiv - AVG_DIV_RANGE * mult;
            bindVars[16] = ref.avgDiv + AVG_DIV_RANGE * mult;
        }
        mult = ((Math.abs(ref.avgCurl)-1) / 4) + 1;
        if (ref.avgCurl == 0) {
            bindVars[17] = -1;
            bindVars[18] = 1;
        } else {
            bindVars[17] = ref.avgCurl - AVG_CURL_RANGE * mult;
            bindVars[18] = ref.avgCurl + AVG_CURL_RANGE * mult;
        }
    }
}

class ClusLinkQuery extends ClusQuery {
    
    final static String LINK_QUERY = 
        "SELECT link_mvclus_id,link_mvclus_wt FROM mvclusterlink" +
        " WHERE mvcluster_id = ?";
    
    protected PreparedStatement linkQuery;
    protected PreparedStatement clusQueryByID;
    
    ClusLinkQuery(Connection dbc) {
        super(dbc);
    }
    
    Clus selectCluster(Clus ref) {
        if (ref.linkWtSum < 1) {
            return null;
        }
        
        if (linkQuery == null) {
            init();
        }
        
        int linkID = -1;
        ResultSet linkRS = null;
        try {
            linkQuery.setInt(1, ref.clusterID);
            linkRS = linkQuery.executeQuery();
            int rand = (int) (Math.random() * ref.linkWtSum);
            while (linkRS.next()) {
                rand -= linkRS.getInt(2);
                if (rand < 0) {
                    break;
                }
            }
            linkID = linkRS.getInt(1);
            linkRS.close();
            linkRS = null;
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (linkRS != null) {
                try { linkRS.close(); } catch (SQLException e) {}
                linkRS = null;
            }
        }
        
        return (linkID > -1) ? getCluster(linkID) : null;
    }
    
    void fillBindVars(Clus ref) {
        // Do nothing.
    }
    
    void close() {
        super.close();
        try {
            if (linkQuery != null) {
                linkQuery.close();
                linkQuery = null;
            }
            if (clusQueryByID != null) {
                clusQueryByID.close();
                clusQueryByID = null;
            }
        } catch (SQLException e) {
             e.printStackTrace();
        } finally {
            if (linkQuery != null) {
                try { linkQuery.close(); } catch (SQLException e) {}
                linkQuery = null;
            }
            if (clusQueryByID != null) {
                try { clusQueryByID.close(); } catch (SQLException e) {}
                clusQueryByID = null;
            }
        }
    }
    
    protected void init() {
        try {
            this.linkQuery = dbc.prepareStatement(LINK_QUERY);
            this.clusQueryByID = dbc.prepareStatement(MVCLUSTER_QUERY_BY_ID);
        } catch (SQLException e) {
            e.printStackTrace();
        }
    }
    
    protected Clus getCluster(int clusID) {
        Clus c = null;
        ResultSet rs = null;
        try {
            clusQueryByID.setInt(1, clusID);
            rs = clusQueryByID.executeQuery();
            if (rs.next()) {
                c = Clus.parseRow(rs);
            }
            rs.close();
            rs = null;
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (rs != null) {
                try { rs.close(); } catch (SQLException e) {}
                rs = null;
            }
        }
        return c;
    }
}

