/*
 * com/ngon/tap/Evol.java
 * Victor Liu > mailto:victor@n-gon.com
 *
 * Copyright (C) 2004 Victor Liu
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version. 
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details. 
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package com.ngon.tap;

import com.ngon.io.TextFileRunner;
import java.io.IOException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * class Evol
 */
class Evol implements Evolver {
    
    final static int RANDOM = 13;
    
    protected String dbHost       = "localhost";
    protected String dbUser       = "victor";
    protected String dbPass       = "";
    
    protected Connection dbc = null;
    protected int[] initialClusters = null;
    protected Hashtable clipInfo = null;
    protected ClusQuery queryInSequence = null;
    protected ClusQuery queryOutOfSequence = null;
    protected PreparedStatement clipQuery = null;
    protected PreparedStatement blocksQuery = null;
    protected PreparedStatement clusQueryByID = null;
    
    public Evol(String dbHost, String dbUser, String dbPass) {
        this.dbHost = dbHost;
        this.dbUser = dbUser;
        this.dbPass = dbPass;
    }
    
    public Evol(Connection dbc) {
        this.dbc = dbc;
    }
    
    public void init() {
        try {
            if (dbc == null) {
                Class.forName("com.mysql.jdbc.Driver");
                String s = "jdbc:mysql://" + dbHost + "/c3cuts?user=" + dbUser;
                if (dbPass != "") {
                    s += "&password=" + dbPass;
                }
                dbc = DriverManager.getConnection(s);
            }
            
            // Prepared queries.
            queryInSequence = new ClusQueryInSequence(dbc);
            queryOutOfSequence = new ClusLinkQuery(dbc);
            // Use the following instead if the mvclusterlink table 
            // has not yet been filled (by running TapLinks).
            // TODO: Program this logic.
//                queryOutOfSequence = new ClusQueryOutOfSequence(dbc);
//                queryOutOfSequence.setVerbose(true);
            
            clipQuery = dbc.prepareStatement(
                "SELECT name,len FROM clip WHERE id = ?");
            blocksQuery = dbc.prepareStatement(
                "SELECT block_nr,mv_x,mv_y FROM mvblock" +
                " WHERE mvcluster_id = ?");
            clusQueryByID = dbc.prepareStatement(
                ClusQuery.MVCLUSTER_QUERY_BY_ID);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    public Hashtable getClipInfo() {
        if (clipInfo == null) {
            clipInfo = Evol.getClipInfo(dbc);
        }
        return clipInfo;
    }
    
    public boolean evol(Taps taps, int maxTaps, ClipSegFactory csFactory) {
        // Loop through queue; for each cluster, select the next one,
        // and update it with the new cluster data.  Exception:
        // for the last item, do an unrestricted search over all
        // clips.  If selection is from a different clip, add it as 
        // a new item to the queue.
        for (int i = 0; i < taps.size(); i++) {
            Tap tap = (Tap) taps.elementAt(i);
            if (tap.clus.clip.isLastFrame()) {
                if (Debug.DEBUG) {
                    System.out.println(
                        "LAST FRAME for clip " + tap.clus.clipID + 
                        "." + tap.clus.frameNr);
                }
                taps.killTap(i);
                continue;
            }
            
            // Advance to next frame in clip, and look for a cluster
            // within the next frame.
            if (tap.clus.clip.nextFrame()) {
                tap.nFailedAdvances = 0;
                Clus next = queryInSequence.selectCluster(tap.clus);
                if (next == null) {
                    // No clusters in next frame; increment cluster's 
                    // frameNr anyway.
                    tap.clus.frameNr++;
                    // Set next in case we branch.
                    next = new Clus();
                    next.copyFields(tap.clus);
                } else {
                    tap.clus.copyFields(next);
                    tap.clus.blocks = getBlocks(tap.clus);
                }
                
                // For last cluster in taps, chance to branch 
                // out of sequence.
                if (i == taps.size()-1 &&
                    (int) (Math.random() * RANDOM) == 0) {
                    Clus branch = queryOutOfSequence.selectCluster(next);
                    if (branch == null) {
                        break;
                    }
                    if (Debug.DEBUG) {
                        System.out.println("Branching " + branch.clipID +
                            "." + branch.frameNr + " from " + next.clipID + 
                            "." + next.frameNr);
                    }
                    branch.clip = csFactory.create(branch.clipID, 
                                                   branch.frameNr);
                    if (branch.clip != null) {
                        branch.clip.start();
                        branch.blocks = getBlocks(branch);
                        taps.addElement(new Tap(branch));
                    }
                }
            } else {
                ++tap.nFailedAdvances;
                // Remove this cluster.
                if (Debug.DEBUG) {
                    System.out.println("FAILED TO ADVANCE FRAME for clip " + 
                        tap.clus.clipID + "." + tap.clus.frameNr + 
                        " fails=" + tap.nFailedAdvances);
                }
                if (tap.nFailedAdvances > NULL_FRAME_RETRIES) {
                    taps.killTap(i);
                }
            }
        }
        
        if (taps.size() > maxTaps) {
            taps.killTap(0);
        } else if (taps.size() == 0) {
            addNewTap(taps, csFactory);
        }
        
        return true;
    }
    
    public void kill() {
        initialClusters = null;
        if (clipInfo != null) {
            clipInfo.clear();
            clipInfo = null;
        }
    }
    
    public Block[] getBlocks(Clus clus) {
        Vector v = new Vector();
        try {
            blocksQuery.setInt(1, clus.clusterID);
            ResultSet rs = blocksQuery.executeQuery();
            while (rs.next()) {
                v.addElement(Block.parseRow(rs));
            }
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        Block[] blocks = new Block[v.size()];
        v.copyInto(blocks);
        return blocks;
    }
    
    public void setInitialClusters(int[] initialClusters) {
        this.initialClusters = initialClusters;
    }
    
    public void setClipInfo(Hashtable clipInfo) {
        this.clipInfo = clipInfo;
    }
    
    protected void addNewTap(Taps taps, ClipSegFactory csFactory) {
        Clus c = selectInitialCluster();
        if (c == null) {
            return;
        }
        // Create a new clip to attach to this cluster.
        c.clip = csFactory.create(c.clipID, c.frameNr);
        if (c.clip != null) {
            c.clip.start();
            // Get blocks of cluster.
            c.blocks = getBlocks(c);
            taps.addElement(new Tap(c));
        }
    }
    
    /*
     * Randomly choose one cluster to start with.
     */
    protected Clus selectInitialCluster() {
        if (initialClusters == null) {
            initialClusters = getInitialClusters(dbc);
        }
        int rand = (int) (Math.random() * initialClusters.length);
        int clusID = initialClusters[rand];
        Clus c = getCluster(clusID);
        if (c == null) {
            if (Debug.DEBUG || Debug.DEBUG_APPLET) {
                System.err.println(
                    "selectInitialCluster(): can't find cluster for id = " +
                    clusID);
            }
        }
        return c;
    }
    
    /*
     * Is there a way to get rid of this repeated code (cf ClusLinkQuery) ?
     * Originally had this as a static method in ClusQuery, but that 
     * didn't play well with pooled connections in Tomcat, since the
     * dbc, and hence the static PreparedStatement, got pulled out from
     * under us.
     */
    protected Clus getCluster(int clusID) {
        Clus c = null;
        try {
            clusQueryByID.setInt(1, clusID);
            ResultSet rs = clusQueryByID.executeQuery();
            if (rs.next()) {
                c = Clus.parseRow(rs);
            }
            rs.close();
            rs = null;
        } catch (SQLException e) {
            e.printStackTrace();
        }
        return c;
    }
    
    /*
     * Clean up resources.  Necessary to release db connection back to pool.
     */
    public void releaseDB() {
        try {
            dbc.close();
            dbc = null;
            queryInSequence.close();
            queryInSequence = null;
            queryOutOfSequence.close();
            queryOutOfSequence = null;
            clipQuery.close();
            clipQuery = null;
            blocksQuery.close();
            blocksQuery = null;
            clusQueryByID.close();
            clusQueryByID = null;
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (dbc != null) {
                try { dbc.close(); } catch (SQLException e) {}
                dbc = null;
            }
            if (queryInSequence != null) {
                queryInSequence.close();
                queryInSequence = null;
            }
            if (queryOutOfSequence != null) {
                queryOutOfSequence.close();
                queryOutOfSequence = null;
            }
            if (clipQuery != null) {
                try { clipQuery.close(); } catch (SQLException e) {}
                clipQuery = null;
            }
            if (blocksQuery != null) {
                try { blocksQuery.close(); } catch (SQLException e) {}
                blocksQuery = null;
            }
            if (clusQueryByID != null) {
                try { clusQueryByID.close(); } catch (SQLException e) {}
                clusQueryByID = null;
            }
        }
    }
    
    static public int[] getInitialClusters(Connection dbc) {
        int[] initialClusters = null;
        Statement stmt = null;
        ResultSet rs = null;
        try {
            Vector v = new Vector();
            stmt = dbc.createStatement();
            rs = stmt.executeQuery(
                "SELECT id FROM mvcluster WHERE n_blocks > 5");
            while (rs.next()) {
                v.addElement(new Integer(rs.getInt(1)));
            }
            initialClusters = new int[v.size()];
            int i = 0;
            for (Enumeration e = v.elements(); e.hasMoreElements(); ) {
                initialClusters[i++] = 
                    ((Integer) e.nextElement()).intValue();
            }
            v.removeAllElements();
            rs.close();
            rs = null;
            stmt.close();
            stmt = null;
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (rs != null) {
                try { rs.close(); } catch (SQLException e) {}
                rs = null;
            }
            if (stmt != null) {
                try { stmt.close(); } catch (SQLException e) {}
                stmt = null;
            }
        }
        return initialClusters;
    }
    
    static public Hashtable getClipInfo(Connection dbc) {
        Hashtable clipInfo = null;
        Statement stmt = null;
        ResultSet rs = null;
        try {
            clipInfo = new Hashtable();
            stmt = dbc.createStatement();
            rs = stmt.executeQuery(
                "SELECT id,name,len FROM clip");
            while (rs.next()) {
                ClipInfo ci = ClipInfo.parseRow(rs);
                clipInfo.put(new Integer(ci.clipID), ci);
            }
            rs.close();
            rs = null;
            stmt.close();
            stmt = null;
        } catch (SQLException e) {
            e.printStackTrace();
        } finally {
            if (rs != null) {
                try { rs.close(); } catch (SQLException e) {}
                rs = null;
            }
            if (stmt != null) {
                try { stmt.close(); } catch (SQLException e) {}
                stmt = null;
            }
        }
        return clipInfo;
    }
}

