#!/usr/bin/perl -w

use FindBin qw($Bin);
chdir $Bin;

use Getopt::Long;

$|= 1;

$DFLT_CLIPS_PATH = '../clips';
$DFLT_N_CLIPSEG_FRAMES = 10;

###############################################################################
# Begin MAIN                                                                  #

&get_args;
&get_clip_dirs;
&create_clips;
exit;

# End of MAIN                                                                 #
###############################################################################

sub get_args {
    my $help;
    &GetOptions('c|clips-path=s' => \$Clips_Path,
                'n|nr-clip-frames=i' => \$N_ClipSeg_Frames,
                'i|iqscale=i' => \$IQScale,
                'p|pqscale=i' => \$PQScale,
                'b|bqscale=i' => \$BQScale,
                'h|help' => \$help);
    (&print_usage && exit) if $help;
    $Clips_Path ||= $DFLT_CLIPS_PATH;
    $N_ClipSeg_Frames ||= $DFLT_N_CLIPSEG_FRAMES;
}

sub print_usage {
    print <<HERE;
$0
Create small clip segments from a set of still image files.

Usage:
  -c, --clips-path      Path to clips
  -n, --nr-clip-frames  Nr of frames per clip segment
  -i, --iqscale         Set Q-factor for I-frames
  -p, --pqscale         Set Q-factor for P-frames
  -b, --bqscale         Set Q-factor for B-frames
HERE
}

sub get_clip_dirs {
    opendir(DIR, $Clips_Path) || die "can't opendir $Clips_Path: $!";
    my @files = readdir(DIR);
    @Clip_Names = grep { (! /^\./) && (-d "$Clips_Path/$_") } @files;
    closedir DIR;
}

##
# For each clip, create temporary copies of the mpeg_encode param files,
# replacing parameters appropriate to a segment of the clip.  Run mpeg_encode 
# and remove the temporary param files.  Output mpegs will be in a subdirectory
# named $N_ClipSeg_Frames.
#
sub create_clips {
    chdir $Clips_Path;
    for my $clip_name (@Clip_Names) {
        my $clip_dir = "$Clips_Path/$clip_name";
        chdir $clip_name;
        mkdir "$N_ClipSeg_Frames" || die $!;
        local $/; # Slurp
        open(PARAMS, "<clip-segs.param") || die $!;
        my $buf = <PARAMS>;
        # Look for frames range.
        unless ($buf =~ /^$clip_name\.\*\.jpg\s+\[(\d+)-(\d+)(\+\d)*\]$/m) {
            die "can't find frames range in file";
        }
        my $nr_digits = length($2);
        my $begin = $1;
        my $end = $2;
        my $incr = $3;
        if ($incr) {
            $incr =~ s/\+//;
        } else {
            $incr = 1;
        }
        unless (defined($begin) && defined($end) && $end > $begin) {
            die "invalid frame range: $begin => $end";
        }
        my $count = 1;
        for (my $i = $begin; 
             $i <= $end - $incr; 
             $i += $N_ClipSeg_Frames * $incr) {
            # Copy $buf.
            my $tmp_buf = $buf;
            my $append = sprintf("%04d", $count);
            my $output_file = "$N_ClipSeg_Frames/$clip_name-$append.mpg";
            my $format = "%0${nr_digits}d";
            my $f1 = sprintf($format, $i);
            my $last_frame = $i + $N_ClipSeg_Frames * $incr - 1;
            if ($last_frame > $end) {
                $last_frame = $end;
            }
            my $f2 = sprintf($format, $last_frame);
            print "Creating $output_file ... ";
            $tmp_buf =~ s|^OUTPUT(\s+)(.+)|OUTPUT$1$output_file|m;
            $tmp_buf =~ s/ ^$clip_name\.\*\.jpg(\s+)\[\d+-\d+(\+*\d*\])$ 
                         /$clip_name.*.jpg${1}[${f1}-${f2}${2}/mx;
            $tmp_buf =~ s/^(IQSCALE\s+)(\d+)$/$1$IQScale/ if $IQScale;
            $tmp_buf =~ s/^(PQSCALE\s+)(\d+)$/$1$PQScale/ if $PQScale;
            $tmp_buf =~ s/^(BQSCALE\s+)(\d+)$/$1$BQScale/ if $BQScale;
            open(TMP_PARAMS, '>tmp.param');
            print TMP_PARAMS $tmp_buf;
            close(TMP_PARAMS);
            # Run mpeg_encode command 
            my $cmd = 'mpeg_encode -realquiet tmp.param';
            system $cmd;
            print "done\n";
            $count++;
        }
        unlink 'tmp.param';
        chdir '..';
    }
}

